
// Parameterized implementation of List ADT
//   with representation classes hidden


object List {

  // creators

  def empty[A] ():List[A] = new ListEmpty[A]()

  def singleton[B] (i:B):List[B] = new ListSingleton[B](i)

  def merge[C] (L:List[C], M:List[C]):List[C] = new ListMerge[C](L,M)


  private class ListEmpty[T] () extends List[T] {

    def isEmpty ():Boolean = true
    def first ():T = throw new RuntimeException("empty().first()")
    def rest ():List[T] = throw new RuntimeException("empty().rest()")

    def length ():Int = 0
/*
    def update (v:T):Unit = throw new RuntimeException("empty().update()")
*/

    def find (v:T):Boolean = 
    //def find[A>:T] (v:A):Boolean = false
      false



    override def hashCode ():Int = 41
    
    override def toString ():String = ""
  }



  private class ListSingleton[U] (i:U) extends List[U] {

    private var value:U = i
    
    def isEmpty ():Boolean = false
    def first ():U = value
    def rest ():List[U] = List.empty()

    def length ():Int = 1
/*
    def update (v:U):Unit = { value = v }
*/

    def find (v:U):Boolean = 
    //def find[A>:U] (v:A):Boolean =
      ( v==value )
    
    override def hashCode ():Int = 41 + i.hashCode()

    override def toString ():String = " " + value.toString() 
  }


  
  private class ListMerge[V] (L:List[V], M:List[V]) extends List[V] {

    def isEmpty ():Boolean = 
      (L.isEmpty() && M.isEmpty())

    def first ():V = 
      if (L.isEmpty())
        M.first()
      else
        L.first()
    
    def rest ():List[V] = 
      if (L.isEmpty())
        M.rest()
      else
        List.merge(L.rest(),M)

    def length ():Int = L.length() + M.length()
/*
    def update (v:V):Unit = 
      if (L.isEmpty())
        M.update(v)
      else
         L.update(v)
*/

    def find (v:V):Boolean = 
    //def find[A>:V] (v:A):Boolean =
      (L.find(v) || M.find(v))


    
    override def hashCode ():Int = 
      41 * (
        41 + L.hashCode()
      ) + M.hashCode()

    override def toString ():String = L.toString() + M.toString()
  }

}


abstract class List[T] {  // <-- makes List covariant in T


  def isEmpty ():Boolean
  def first ():T
  def rest ():List[T]
  def length ():Int

  // def update (v:T):Unit   // mutates the first element of a list

  def find (v:T):Boolean

  // def find[A>:T] (v:A):Boolean

}

